from PIL import Image
import numpy as np
import tensorflow as tf
import uuid
import json
import datetime
import os
import time

from ..general.general import mkdir_p

def write_json(new_data, filename):
    if not os.path.isfile(filename):
        file = open(filename, 'w+')
        file.write(json.dumps({"results": []}, indent=1))
        file.close()
        time.sleep(0.1)

    file = open(filename, 'r')
    content = file.read()
    file.close()

    file_data = json.loads(content)
    file_data['results'].append(new_data)

    file = open(filename, 'w')
    file.write(json.dumps(file_data, indent=1))
    file.close()


def save_image(id, original_image, classifier, model_name, result):
    if bool(classifier['config_record']['save_images']):
        mkdir_p(f"{classifier['config_record']['model_log_directory']}\\{result['class']}")
        original_image.save(f"{classifier['config_record']['model_log_directory']}\\{result['class']}\\{id}.png")
        write_json(result, f"{classifier['config_record']['model_log_directory']}\\image_data.json")


def predict(original_image, classifier, model_name):
    img = original_image.resize(
        (classifier['meta']['image_resize'][1],
         classifier['meta']['image_resize'][0]),
        resample=Image.Resampling.NEAREST)

    img_array = tf.keras.utils.img_to_array(img)
    img_array = tf.expand_dims(img_array, 0)  # Create a batch

    id = str(uuid.uuid1())
    predictions = classifier['model'].predict(img_array)
    scores = tf.nn.softmax(predictions[0])
    result = {
        'epoc_time': str(datetime.datetime.utcnow().timestamp() * 1000),
        'argmax_index': int(np.argmax(scores)),
        'value_at_argmax': str(scores[np.argmax(scores)].numpy()),
        'class': classifier['meta']['classes'][np.argmax(scores)],
        'classes': classifier['meta']['classes'],
        'scores': scores.numpy().tolist(),
        'id': id,
        'image_saved': 0,
        'model_name': model_name
    }

    save_image(id, original_image, classifier, model_name, result)

    return result



import numpy as np
import tensorflow as tf
import pathlib

from PIL import Image, ImageDraw

from tensorflow.keras import layers
from tensorflow.keras.models import Sequential
from sklearn.metrics import confusion_matrix
from sklearn.utils.class_weight import compute_class_weight

def build_and_train(root_image_directory, model_location,
                    model_name, epochs=10,
                    resize_ratio=0.2, auto_balance_data=True):
    data_dir = pathlib.Path(root_image_directory)

    image_list = list(data_dir.glob('*/*.png'))
    image_count = len(image_list)
    print(image_count)

    img = Image.open(image_list[0])

    batch_size = 1
    img_height = int(img.height * resize_ratio)
    img_width = int(img.width * resize_ratio)

    train_ds = tf.keras.utils.image_dataset_from_directory(
        data_dir,
        validation_split=0.2,
        subset="training",
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size)

    val_ds = tf.keras.utils.image_dataset_from_directory(
        data_dir,
        validation_split=0.2,
        subset="validation",
        seed=123,
        image_size=(img_height, img_width),
        batch_size=batch_size)

    class_names = train_ds.class_names
    print(f"Class Names: {class_names}")

    class_weights = None
    if auto_balance_data:
        y_train = np.concatenate([y for x, y in train_ds], axis=0)

        class_weights = compute_class_weight(class_weight="balanced", classes=np.unique(y_train), y=y_train)
        class_weights = {i: w for i, w in enumerate(class_weights)}

        print(f"Class Weights: {class_weights}")




    AUTOTUNE = tf.data.AUTOTUNE

    train_ds = train_ds.cache().shuffle(1000).prefetch(buffer_size=AUTOTUNE)
    val_ds = val_ds.cache().prefetch(buffer_size=AUTOTUNE)

    num_classes = len(class_names)

    model = Sequential([
        layers.Rescaling(1. / 255, input_shape=(img_height, img_width, 3)),
        layers.Conv2D(16, 3, padding='same', activation='relu'),
        layers.MaxPooling2D(),
        layers.Conv2D(32, 3, padding='same', activation='relu'),
        layers.MaxPooling2D(),
        layers.Conv2D(64, 3, padding='same', activation='relu'),
        layers.MaxPooling2D(),
        layers.Flatten(),
        layers.Dense(128, activation='relu'),
        layers.Dropout(0.5),
        layers.Dense(num_classes)
    ])

    model.compile(optimizer='adam',
                  loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                  metrics=['accuracy'])

    history = model.fit(
        train_ds,
        validation_data=val_ds,
        epochs=epochs,
        class_weight=class_weights
    )
    train_data = list(train_ds)
    features = np.concatenate([train_data[n][0] for n in range(0, len(train_data))])
    targets = np.concatenate([train_data[n][1] for n in range(0, len(train_data))])
    print(targets)
    predictions = model.predict(features)
    print(predictions.argmax(1))

    cf = confusion_matrix(targets, predictions.argmax(1).astype(int))

    model.save(model_location)

    stats = {
        'cm': cf
    }
    rendering = {
        'image_resize': [img_height, img_width],
        'classes': class_names
    }

    return stats, rendering



from loguru import logger


from .model_management.model_manager import load_models_from_config
from .prediction import prediction


class universal_predictor:
    def __new__(cls, **args):  # make singleton
        if not hasattr(cls, 'instance'):
            cls.instance = super(universal_predictor, cls).__new__(cls)
            cls.instance.__initialized = False
        return cls.instance

    # logger available for dependency injection, no documentation why this is inappropriate.
    def __init__(self, config=None, verbose=0):
        if (self.__initialized): return  # make singleton
        self.__initialized = True
        self.verbose = verbose
        self.config = config
        self.classifiers = {}

        # region ----- validation
        if self.config is None:
            error = "universal_predictor missing configuration"
            logger.error(error)
            raise Exception(error)
        # endregion

        self.classifiers = {}
        self.regressors = {}

        self.load_models()
        logger.debug('universal_predictor initialized')

    # region ----- load
    def load_models(self):
        self.classifiers = load_models_from_config(self.config)
        return None

    # endregion

    # region ----- prediction
    def predict(self, original_image, model_name):
        result = prediction.predict(original_image, self.classifiers[model_name], model_name)
        return result
    # endregion



import errno
import json
import os
import requests
import xmltodict
from loguru import logger


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def get_gcp_config(config_record):
    r = requests.get(config_record['download_latest_from'], allow_redirects=True)
    data_dict = xmltodict.parse(r.content)
    available_models = {}
    for obj in data_dict['ListBucketResult']['Contents']:
        key_split = obj['Key'].split('/')
        if config_record['model_name'] == key_split[0]:
            available_models[key_split[0]] = available_models[key_split[0]] if key_split[0] in available_models else {}
            available_models[key_split[0]][key_split[1]] = obj

    logger.info(f"GCP Downloaded for {config_record['model_name']}, {len(available_models.keys())} records found")

    if len(available_models.keys()) == 0:
        raise Exception("No Models Configured for Download")

    return available_models


def load_installed_model(config_record):
    f = open(f"{config_record['model_root_directory']}\\{config_record['model_name']}_gcp.json", "r")
    installed_models = json.loads(f.read())
    f.close()
    return installed_models


def sync_model(config_record):
    available_models = get_gcp_config(config_record)
    installed_models = load_installed_model(config_record)
    if installed_models[config_record['model_name']] != available_models[config_record['model_name']]:
        logger.info(f"Update available, downloading...")
        download_model(config_record, available_models)
    else:
        logger.info(f"Models up to date")


def download_model(config_record, available_models=None):
    if available_models == None:
        available_models = get_gcp_config(config_record)

    for file_key in available_models[config_record['model_name']]:
        logger.info(f"Downloading {config_record['model_root_directory']}\\{file_key}")
        outfile = f"{config_record['model_root_directory']}\\{file_key}"
        mkdir_p(config_record['model_root_directory'])
        upfile = f"{config_record['download_latest_from']}{available_models[config_record['model_name']][file_key]['Key']}"
        r = requests.get(upfile, allow_redirects=True, stream=True)
        open(outfile, 'wb').write(r.content)

    logger.info(f"Downloading {config_record['model_name']}_gcp.json")
    outfile = f"{config_record['model_root_directory']}\\{config_record['model_name']}_gcp.json"
    open(outfile, 'w').write(json.dumps(available_models, indent=1))

    logger.info(f"Download complete")



import json
import os
import errno

import tensorflow as tf
from loguru import logger

from .download_models import download_model, sync_model
from ..general.general import mkdir_p


def construct_dirs(config_record):
    if not os.path.isdir(config_record['model_root_directory']):
        mkdir_p(config_record['model_root_directory'])

    if not os.path.isdir(config_record['model_log_directory']):
        mkdir_p(config_record['model_log_directory'])


def get_local_file_locations(config_record):
    meta_file = f"{config_record['model_root_directory']}\\{config_record['model_name']}_meta.json"
    model_file = f"{config_record['model_root_directory']}\\{config_record['model_name']}_model.h5"
    gcp_file = f"{config_record['model_root_directory']}\\{config_record['model_name']}_gcp.json"

    meta_file_exist = os.path.isfile(meta_file)
    model_file_exist = os.path.isfile(model_file)
    gcp_file_exist = os.path.isfile(gcp_file)

    return meta_file, model_file, gcp_file, meta_file_exist, model_file_exist, gcp_file_exist


def process_model_config(config):
    if 'public_models' in config:
        for config_record in config['public_models']:
            construct_dirs(config_record)
            ingest_public_model(config_record)
            logger.info(f"PreProcessed public_models {config_record['model_name']}")
            pass
    # TODO, Build This
    if 'private_models' in config:
        for config_record in config['private_models']:
            ingest_private_model(config_record)
            logger.info(f"PreProcessed private_models {config_record['model_name']}")
            pass


def load_models_from_config(config):
    process_model_config(config)
    classifiers = {}
    if 'public_models' in config:
        for config_record in config['public_models']:
            classifiers[config_record['model_name']] = load_model(config_record)
            logger.info(f"Loaded {config_record['model_name']}")
            pass
    # TODO, Build This
    if 'private_models' in config:
        for config_record in config['private_models']:
            pass
    return classifiers


def load_model(config_record):
    meta_file, model_file, gcp_file, meta_file_exist, model_file_exist, gcp_file_exist = get_local_file_locations(
        config_record)
    # TODO, make thi agnostic to model type

    if not meta_file_exist or not model_file_exist or not gcp_file_exist:
        error = 'Attempted to load models that are not downloaded'
        logger.error(error)
        raise Exception(error)

    classifier = {
        'config_record': config_record,
        'model': tf.keras.models.load_model(model_file),
        'meta': json.loads(open(meta_file, 'r').read()),
        'gcp': json.loads(open(gcp_file, 'r').read()),
    }

    return classifier


def ingest_public_model(config_record):
    meta_file, model_file, gcp_file, meta_file_exist, model_file_exist, gcp_file_exist = get_local_file_locations(
        config_record)

    locals_exists = False
    if meta_file_exist and model_file_exist and gcp_file_exist:
        locals_exists = True

    if not locals_exists:
        if bool(config_record['download_latest']):
            # Download Latest regardless of existence
            download_model(config_record)
            pass
        else:
            # No model exists and we cant download? but we are asking to run this model?
            error = f"{config_record['model_name']} does not completely exist, not permitted to download fresh copy."
            error += f" - Meta file exists at '{meta_file}':{meta_file_exist}"
            error += f" - Model file exists at '{model_file}':{model_file_exist}"
            error += f" - GCP file exists at '{gcp_file}':{gcp_file_exist}"
            logger.error(error)
            raise Exception(error)
        # download everything and finished
        pass
    else:
        # validate latest and download if allowed
        if bool(config_record['download_latest']):
            # compare GCPs and download if needed
            sync_model(config_record)
            pass
        else:
            # do nothing, we are done
            pass

    try:

        f = open(meta_file, "r")
        installed_models = json.loads(f.read())
        f.close()
    except:
        logger.info('installed_models file missing, proceeding w/ new.')
        pass
    return


def ingest_private_model(config_record):
    # TODO, Build this
    return
